#!/QOpenSys/pkgs/bin/perl

use Getopt::Long;

use strict;

my $func;
my $buildroot = "";


Getopt::Long::Configure ("no_ignore_case");

GetOptions("requires|R" => \&provide_or_requires,
           "provides|P" => \&provide_or_requires,
           "buildroot=s" => \$buildroot)
  or die("Error in command line arguments\n");

die "You must specify either --provides or --requires\n" if (not defined $func);

my %printed;

foreach (<STDIN>) {
    chomp;
    
    my $filename = $_;
    
    my @st = stat $filename;
    my $file_size = $st[7];

    my $rc = open my $fh, '<:raw', $filename;
    
    if(!$rc) {
        warn("Couldn't open $filename: $!");
        next;
    }
    
    my @headers;

    my $header = parse_xcoff($fh, 0, $file_size);
    
    if ($header) {
        push @headers, {
            'header' => $header
        };
    } else {
        seek $fh, 0, 0;
        @headers = parse_archive($fh, 0, $file_size);
    }
    close $fh;
    
    foreach $header (@headers) {
        $func->($header, $filename, $buildroot);
    }
}

exit(0);

sub provide_or_requires
{
    my ($opt, $val) = @_;

    if (defined $func) {
        die("Can't specify --provides and --requires");
    }
    
    $func = ($opt eq 'requires') ? \&requires : \&provides;
}



sub provides
{
    my ($header, $filename) = @_;
    
    $filename =~ s|$buildroot||;
    
    my $line;
    
    if ($header->{header}->{shared}) {
        $line = "lib:$filename";
        
        if ($header->{file}) {
            $line .= "($header->{file})"
        }
        
        $line .= "($header->{header}->{arch})";
    }
    
    if (! exists $printed{$line}) {
        print "$line\n";
        $printed{$line} = 1;
    }
}

sub requires
{
    my ($header, $filename, $buildroot) = @_;
    
    # extract LIBPATH to array, ignoring any non-fully qualified paths
    my @dirs = grep { $_ =~ /^\/.*/ } split /:/, $header->{header}->{libpath};

    @dirs = map to_pase($_), @dirs;
    my $line = '';
    
    foreach my $dep (@{$header->{header}->{deps}}) {
        my $path;
        
        # Ignore /unix, ., and .. dependencies:
        # /unix is provided by the kernel
        # . and .. are provided by the run time linker: https://www.ibm.com/support/knowledgecenter/en/ssw_aix_71/com.ibm.aix.genprogc/shared_object_runtime_linking.htm
        if ("$dep->{path}$dep->{library}" eq "/unix" or
            "$dep->{path}$dep->{library}" eq "." or
            "$dep->{path}$dep->{library}" eq "..") {
            next;
        }
        
        if ($dep->{path}) {
            $path = to_pase($dep->{path});
        } else {
            foreach my $d(@dirs) {
                if (-e "$d/$dep->{library}" || -e "$buildroot/$d/$dep->{library}") {
                    $path = $d;
                    last;
                }
            }
        
            if (! defined $path) {
                $path = @dirs[0];
                warn("$filename: Can't find directory for $dep->{library}\n");
            }
        }
        
        $line = "lib:$path/";
        
        warn "$filename missing library\n" if(not $dep->{library});
        $line .= $dep->{library};
        
        if ($dep->{member}) {
            $line .= "($dep->{member})";
        }
        
        $line .= "($header->{header}->{arch})";
        
        if (! exists $printed{$line}) {
            print "$line\n";
            $printed{$line} = 1;
        }
    }
}


sub parse_xcoff
{
    # XCOFF format documented here: https://www.ibm.com/support/knowledgecenter/en/ssw_aix_72/com.ibm.aix.files/XCOFF.htm
    my ($fh, $off_start, $off_end) = @_;
    
    # Read magic bytes and number of sections from the XCOFF header
    my $bytes_read = read $fh, my $bytes, 2;
    my $magic = unpack 'n', $bytes;

#     printf "%04X\n", $magic;
    if ($magic != 0x01DF and $magic != 0x01F7) {
        return 0;
    }
    
    my $is64 = $magic == 0x01F7;
    
    read $fh, $bytes, 2;
    my $f_nscns = unpack 'n', $bytes;

    # Read the size of the auxiliary header so we can skip past it to the first section header
    seek $fh, $off_start+16, 0;
    $bytes_read = read $fh, $bytes, 4;

    my ($f_opthdr, $f_flags) = unpack 'nn', $bytes;
    
    my $is_shrobj = $f_flags & 0x2000;

    # size of the xcoff file header
    my $file_header_size = $is64 ? 24 : 20;

    my $offset = $file_header_size + $f_opthdr;
    seek $fh, $off_start+$offset, 0;

    my $found = 0;
    for(my $i = 0; $i < $f_nscns; ++$i)
    {
        $bytes_read = read $fh, $bytes, 8;
        my $section = unpack 'Z8', $bytes;

        if ($section eq '.loader')
        {
            $found = 1;
            last;
        }
        
        $offset += $is64 ? 72 : 40;
        seek $fh, $off_start+$offset, 0;
    }

    return 0 if (not $found);

    # Read the offset to the loader section raw data
    my $s_scnptr_offset = $is64 ? 32 : 20;
    $offset += $s_scnptr_offset;
    seek $fh, $off_start+$offset, 0;
    $bytes_read = read $fh, $bytes, ($is64 ? 8 : 4);

    my $s_scnptr_fmt = ($is64 ? 'Q>' : 'N');
    my $s_scnptr = unpack $s_scnptr_fmt, $bytes;

    $offset = $s_scnptr;

    # Read the import file id string table length and # of import file ids
    seek $fh, $off_start+$s_scnptr+12, 0;
    $bytes_read = read $fh, $bytes, 8;
    my ($l_istlen, $l_nimpid) = unpack 'NN', $bytes;

    # Read the offset to the start of the import file ids
    my $l_impoff_offset = $is64 ? 24 : 20;
    $offset = $s_scnptr + $l_impoff_offset;
    seek $fh, $off_start+$offset, 0;
    $bytes_read = read $fh, $bytes, $is64 ? 8 : 4;
    my $l_impoff_fmt = ($is64 ? 'Q>' : 'N');
    my $l_impoff = unpack $l_impoff_fmt, $bytes;

    # Read the whole file import id string table
    $offset = $s_scnptr + $l_impoff;
    seek $fh, $off_start+$offset, 0;
    $bytes_read = read $fh, $bytes, $l_istlen;


    # Unpack the PATH and calculate the difference in length
    my $path;
    my @deps;

    for (my $idx = 0; $idx < $l_nimpid; $idx++) {
        my @tokens;
        
        for (my $i = 0; $i < 3; $i++) {
            my $str = unpack "Z$l_istlen", $bytes;
            my $str_len = (length $str) + 1;
                        
            $l_istlen -= $str_len;
            $bytes = substr $bytes, $str_len;
            
            push @tokens, $str;
        }
        
        # First entry is always the embeded LIBPATH
        if ($idx == 0) {
            $path = @tokens[0];
        } else {
            push @deps, {
                'path'    => @tokens[0],
                'library' => @tokens[1],
                'member'  => @tokens[2]
            };
        }
    }
    
    return { 'libpath' => $path, 'deps' => \@deps, 
             'arch' => ($is64 ? "ppc64" : "ppc"),
             'shared' => $is_shrobj
    };
}

sub parse_archive {
    my ($fh) = @_;
    
    my $bytes;
    my $bytes_read;
    
    read $fh, $bytes, 8;
    if ($bytes != "<bigaf>\n") {
        return 0;
    }
    
    read $fh, my $fl_memoff, 20;
    read $fh, $bytes, 40;
    read $fh, my $fl_fstmoff, 20;
    read $fh, my $fl_lstmoff, 20;
    
    my $offset = $fl_fstmoff;
    
    my @headers;
    
    while(1) {
        seek $fh, $offset, 0;
        
        read $fh, my $ar_size, 20;
        read $fh, my $ar_nxtmem, 20;
        read $fh, my $ar_prvmem, 20;
        read $fh, my $ar_date, 12;
        read $fh, my $ar_uid, 12;
        read $fh, my $ar_gid, 12;
        read $fh, my $ar_mode, 12;
        read $fh, my $ar_namlen, 4;
        
        read $fh, my $ar_name, $ar_namlen;
        
        my $curoffset = tell $fh;
        $curoffset += 2; # Skip past ar_fmag
        
        if ($curoffset % 2 == 1) {
            $curoffset += 1;
        }
        seek $fh, $curoffset, 0;
        
        my $info = parse_xcoff($fh, $curoffset, $curoffset+$ar_size);
        if ($info) {
            push @headers, { 
                'file' => $ar_name,
                'header' => $info
            };
        }
        
        if ($offset == $fl_lstmoff) {
            last;
        } else {
            $offset = $ar_nxtmem;
        }
    }
    
    return @headers;
}

# map AIX library paths to PASE paths
sub to_pase {
    my $path = shift;
    
    my %paths;
    @paths{qw(/lib /usr/lib /usr/vac/lib /usr/vacpp/lib)} = ();
    
    if (exists $paths{$path}) {
        $path = '/QOpenSys/usr/lib';
    }
    
    return $path;
}
